from __future__ import annotations
from nestful.schemas.openapi import Component
from pydantic import BaseModel, ConfigDict, model_validator
from typing import List, Dict, Optional, Union


class QueryParameter(BaseModel):
    type: Optional[str] = None
    description: Optional[str] = None
    required: bool = False
    enum: List[str] = []
    # TODO: https://github.com/TathagataChakraborti/NESTFUL/issues/2
    allowed_values: Union[str, List[str]] = []
    possible_values: Union[str, List[str]] = []
    default_value: Optional[str] = None
    ################################################################


class MinifiedAPI(BaseModel):
    name: str
    inputs: List[str]
    outputs: List[str]


class API(BaseModel):
    model_config = ConfigDict(extra="ignore")

    name: str
    description: str
    # TODO: https://github.com/TathagataChakraborti/NESTFUL/issues/2
    parameters: Dict[str, QueryParameter] = dict()
    query_parameters: Dict[str, QueryParameter] = dict()
    path_parameters: Dict[str, QueryParameter] = dict()
    arguments: Dict[str, QueryParameter] = dict()
    ################################################################
    output_parameters: Dict[str, Component] = dict()

    @model_validator(mode="after")
    def temporary_field_jarl(self) -> API:
        if self.path_parameters:
            self.query_parameters = self.path_parameters

        if self.parameters:
            self.query_parameters = self.parameters

        if self.arguments:
            self.query_parameters = self.arguments

        return self

    def __str__(self) -> str:
        return str(
            self.dict(
                include={
                    "name",
                    "description",
                    "query_parameters",
                    "output_parameters",
                }
            )
        )

    def get_arguments(self, required: Optional[bool] = True) -> List[str]:
        if required is None:
            return list(self.query_parameters.keys())
        else:
            return [
                key
                for key in self.query_parameters.keys()
                if self.query_parameters[key].required is required
            ]

    def get_outputs(self) -> List[str]:
        outputs = []

        for item in self.output_parameters:
            outputs.append(item)

            if self.output_parameters[item].properties:
                for inner_item in self.output_parameters[item].properties:
                    outputs.append(f"{item}.{inner_item}")

        return outputs

    def minified(self, required: Optional[bool] = True) -> MinifiedAPI:
        return MinifiedAPI(
            name=self.name,
            inputs=self.get_arguments(required),
            outputs=self.get_outputs(),
        )


class Catalog(BaseModel):
    apis: List[API] = []

    def get_api(
        self,
        name: str,
        minified: bool = False,
        required: Optional[bool] = False,
    ) -> Union[API, MinifiedAPI, None]:
        api_object = next(filter(lambda x: x.name == name, self.apis), None)

        if api_object:
            return api_object if not minified else api_object.minified(required)

        return None



from nestful.utils import extract_label, TOKEN
from nestful.schemas.errors import ErrorType
from random import sample, randint
from typing import Optional, Dict, Any, Tuple, List, Set
from copy import deepcopy
from nestful.errors.error_tagger import tag_sequence_step, tag_sequence
from nestful import (
    SequenceStep,
    SequencingData,
    SequencingDataset,
    Catalog,
    AtomicCall,
)

MAX_COLLISIONS = 100


def induce_error_in_step(
    step: SequenceStep,
    catalog: Catalog,
    memory: Dict[str, Any],
    error_type: ErrorType = ErrorType.UNKNOWN,
    num_errors: int = 1,
    referred_only: bool = True,
) -> Tuple[Optional[SequenceStep], Dict[str, Any]]:
    if error_type == ErrorType.UNKNOWN:
        error_type = ErrorType.get_random_error()

    if error_type == ErrorType.MISSING_PARAMETER:
        error_step = remove_required_argument(
            step, catalog, num_errors, referred_only
        )

    elif error_type == ErrorType.MADE_UP_PARAMETER:
        error_step = rename_required_argument(
            step, catalog, num_errors, referred_only
        )

    elif error_type == ErrorType.MISSING_MEMORY:
        new_memory = remove_memory_item(step, memory, num_errors)
        step = tag_sequence_step(step, step, memory)

        return step if new_memory != memory else None, memory

    elif error_type == ErrorType.MADE_UP_ASSIGNMENT:
        error_step = rename_assignment(step, num_errors)

    else:
        raise NotImplementedError(f"Error type {error_type} not supported yet.")

    if error_step:
        error_step = tag_sequence_step(error_step, step, memory)

    return error_step, memory


def induce_error_in_sequence(
    sequence: SequencingData,
    catalog: Catalog,
    memory: Dict[str, Any],
    error_type: ErrorType = ErrorType.UNKNOWN,
    num_errors: int = 1,
    referred_only: bool = True,
) -> SequencingData:
    error_count = 0
    new_sequence = deepcopy(sequence)

    if error_type in [ErrorType.NEW_CALL, ErrorType.MADE_UP_API]:
        raise NotImplementedError(f"Error type {error_type} not supported yet.")

    while error_count < num_errors:
        index = randint(a=0, b=len(new_sequence.output) - 1)
        step = new_sequence.output[index]

        if error_type == ErrorType.MISSING_CALL:
            who_used = new_sequence.who_used(step.label or "")

            if who_used:
                new_sequence = new_sequence.remove_reference(step.label or "")
                error_count += 1
            else:
                continue

        elif error_type == ErrorType.BAD_REPEAT:
            error_step, new_memory = induce_error_in_step(
                step, catalog, memory, referred_only=referred_only
            )

            if error_step is not None:
                new_sequence.output = (
                    new_sequence.output[:index]
                    + [error_step]
                    + new_sequence.output[index:]
                )
                memory = new_memory

                error_count += 1
        else:
            error_step, new_memory = induce_error_in_step(
                step, catalog, memory, error_type, referred_only=referred_only
            )

            if error_step is not None:
                new_sequence.output[index] = error_step
                memory = new_memory

                error_count += 1

    new_sequence = tag_sequence(new_sequence, sequence, memory, catalog)
    return new_sequence


def batch_generate_error_steps(
    dataset: SequencingDataset,
    catalog: Catalog,
    num_samples: int,
    error_type: ErrorType = ErrorType.UNKNOWN,
    num_error_per_sample: int = 1,
    referred_only: bool = True,
    forbidden_indices: Optional[List[int]] = None,
) -> List[AtomicCall]:
    current_samples: List[AtomicCall] = []
    stored_hashes = set()
    total_collisions = 0

    new_dataset = SequencingDataset(data=[])

    if forbidden_indices:
        for index, data in enumerate(dataset.data):
            if index not in forbidden_indices:
                new_dataset.data.append(data)
    else:
        new_dataset.data = dataset.data

    while len(current_samples) < num_samples:
        num_collisions = 0

        while num_collisions < MAX_COLLISIONS:
            random_index = randint(a=0, b=len(new_dataset.data) - 1)
            random_sequence = new_dataset.data[random_index]

            random_index = randint(a=0, b=len(random_sequence.output) - 1)
            step = random_sequence.output[random_index]

            memory = random_sequence.generate_dummy_output(
                catalog, index=random_index
            )

            error_step, new_memory = induce_error_in_step(
                step,
                catalog,
                memory,
                error_type,
                num_error_per_sample,
                referred_only,
            )

            if error_step is not None:
                call_str = error_step.pretty_print(collapse_maps=True)
                new_hash = hash(call_str)

                if new_hash in stored_hashes:
                    num_collisions += 1
                else:
                    stored_hashes.add(new_hash)

                    current_samples.append(
                        AtomicCall(
                            call=error_step,
                            memory=new_memory,
                            ground_truth=AtomicCall(call=step, memory=memory),
                        )
                    )

            if num_collisions == MAX_COLLISIONS:
                total_collisions += 1

                if total_collisions == MAX_COLLISIONS:
                    return current_samples

    return current_samples


def remove_required_argument(
    step: SequenceStep,
    catalog: Catalog,
    num: int = 1,
    referred_only: bool = True,
) -> Optional[SequenceStep]:
    error_step = deepcopy(step)
    required_params = error_step.get_required_args(catalog)

    if referred_only:
        referred_assignments = get_args_with_labeled_assignments(step.arguments)
        required_params = {
            item for item in required_params if item in referred_assignments
        }

    if num > len(required_params):
        return None

    else:
        params_to_remove = sample(list(required_params), num)

        for item in params_to_remove:
            del error_step.arguments[item]

        return error_step


def rename_required_argument(
    step: SequenceStep,
    catalog: Catalog,
    num: int = 1,
    referred_only: bool = True,
) -> Optional[SequenceStep]:
    error_step = deepcopy(step)
    required_params = error_step.get_required_args(catalog)

    if referred_only:
        referred_assignments = get_args_with_labeled_assignments(step.arguments)
        required_params = {
            item for item in required_params if item in referred_assignments
        }

    if num > len(required_params):
        return None

    else:
        params_to_rename = sample(list(required_params), num)

        for item in params_to_rename:
            new_argument = transform_variable(item)
            error_step.arguments[new_argument] = step.arguments[item]

            del error_step.arguments[item]

        return error_step


def remove_memory_item(
    step: SequenceStep, memory: Dict[str, Any], num: int = 1
) -> Dict[str, Any]:
    keys_of_interest = get_labels_from_labeled_assignments(step.arguments)

    if num > len(keys_of_interest):
        return memory

    else:
        key_to_remove = sample(list(keys_of_interest), num)

        for item in key_to_remove:
            memory[item] = {}

        return memory


def rename_assignment(
    step: SequenceStep, num: int = 1
) -> Optional[SequenceStep]:
    error_step = deepcopy(step)
    args_of_interest = get_args_with_labeled_assignments(step.arguments)

    if num > len(args_of_interest):
        return None

    else:
        assignments_to_reassign = sample(list(args_of_interest), num)

        for item in assignments_to_reassign:
            label, mapping = extract_label(str(step.arguments[item]))

            if mapping:
                mapping_components = mapping.split(".")
                mapping_components[-1] = transform_variable(
                    mapping_components[-1]
                )

                new_mapping = ".".join(mapping_components)
                error_step.arguments[item] = f"${label}.{new_mapping}$"

        return error_step


def get_labels_from_labeled_assignments(arguments: Dict[str, Any]) -> Set[str]:
    args_of_interest = set()

    for arg, value in arguments.items():
        label, mapping = extract_label(str(value))

        if label.startswith(TOKEN):
            args_of_interest.add(label)

    return args_of_interest


def get_args_with_labeled_assignments(arguments: Dict[str, Any]) -> Set[str]:
    args_of_interest = set()

    for arg, value in arguments.items():
        label, mapping = extract_label(str(value))

        if label.startswith(TOKEN) and mapping is not None:
            args_of_interest.add(arg)

    return args_of_interest


def transform_variable(name: str) -> str:
    # TODO: ISS24 Need to replace with better generator
    return name[::-1]



from nestful.schemas.api import Catalog, API, MinifiedAPI, QueryParameter
from nestful.schemas.sequences import (
    SequencingDataset,
    SequencingData,
    SequenceStep,
    AtomicCall,
)

from nestful.schemas.openapi import OpenAPI



from nestful import SequenceStep, SequencingData, Catalog
from nestful.schemas.sequences import ErrorTag
from nestful.schemas.errors import ErrorType
from nestful.utils import extract_label, TOKEN
from typing import Dict, Any


def tag_sequence_step(
    step: SequenceStep, ground_truth: SequenceStep, memory: Dict[str, Any]
) -> SequenceStep:
    step.errors = []

    if step.name != ground_truth.name:
        step.errors.append(
            ErrorTag(
                error_type=ErrorType.MADE_UP_API,
                info=step.name,
            )
        )

        return step

    for arg in ground_truth.arguments:
        if arg not in step.arguments:
            step.errors.append(
                ErrorTag(
                    error_type=ErrorType.MISSING_PARAMETER,
                    info=arg,
                )
            )

    for arg, value in step.arguments.items():
        if arg not in ground_truth.arguments:
            step.errors.append(
                ErrorTag(
                    error_type=ErrorType.MADE_UP_PARAMETER,
                    info=arg,
                )
            )

        else:
            original_assignment = ground_truth.arguments[arg]

            if value != original_assignment:
                step.errors.append(
                    ErrorTag(
                        error_type=ErrorType.WRONG_ASSIGNMENT,
                        info={arg: value},
                    )
                )

        label, mapping = extract_label(str(value))

        if label.startswith(TOKEN):
            reference = memory.get(label, {})

            if mapping:
                keys = mapping.split(".")

                for item in keys:
                    if item not in reference:
                        step.errors.append(
                            ErrorTag(
                                error_type=ErrorType.MADE_UP_ASSIGNMENT,
                                info=item,
                            )
                        )

                    reference = reference.get(item, {})

                if not reference:
                    step.errors.append(
                        ErrorTag(
                            error_type=ErrorType.MISSING_MEMORY,
                            info=value,
                        )
                    )

    return step


def tag_sequence(
    sequence: SequencingData,
    ground_truth: SequencingData,
    memory: Dict[str, Any],
    catalog: Catalog,
) -> SequencingData:
    for index, step in enumerate(sequence.output):
        indices_of_interest = [
            i for i, x in enumerate(sequence.output) if x.name == step.name
        ]

        repeat_index = indices_of_interest.index(index)
        target_indices = [
            i for i, x in enumerate(ground_truth.output) if x.name == step.name
        ]

        if repeat_index < len(target_indices):
            target_index = target_indices[repeat_index]
            tmp_memory = sequence.generate_dummy_output(
                catalog=catalog, index=index
            )

            tmp_memory = {**memory, **tmp_memory}

            sequence.output[index] = tag_sequence_step(
                step,
                ground_truth=ground_truth.output[target_index],
                memory=tmp_memory,
            )

        else:
            if len(target_indices) == 0:
                reference_api = catalog.get_api(step.name or "")

                if reference_api is None:
                    sequence.errors.append(
                        ErrorTag(
                            error_type=ErrorType.MADE_UP_API, info=step.name
                        )
                    )

                sequence.errors.append(
                    ErrorTag(error_type=ErrorType.NEW_CALL, info=step.name)
                )

            else:
                sequence.errors.append(
                    ErrorTag(error_type=ErrorType.BAD_REPEAT, info=step.name)
                )

    for step in ground_truth.output:
        name, repeat_index = ground_truth.who_produced(step.label or "")
        label = None if name is None else sequence.get_label(name, repeat_index)

        if label is None or step.name not in [
            item.name for item in sequence.output
        ]:
            sequence.errors.append(
                ErrorTag(error_type=ErrorType.MISSING_CALL, info=step.name)
            )

    return sequence



from nestful.errors.error_generator import (
    induce_error_in_step,
    induce_error_in_sequence,
    batch_generate_error_steps,
)

from nestful.errors.error_tagger import tag_sequence, tag_sequence_step


__all__ = [
    "induce_error_in_step",
    "induce_error_in_sequence",
    "batch_generate_error_steps",
    "tag_sequence",
    "tag_sequence_step",
]



from __future__ import annotations
from enum import StrEnum, auto
from random import choice


class ErrorType(StrEnum):
    MISSING_PARAMETER = auto()
    MADE_UP_PARAMETER = auto()
    MISSING_MEMORY = auto()
    MADE_UP_ASSIGNMENT = auto()
    WRONG_ASSIGNMENT = auto()
    MADE_UP_API = auto()
    BAD_REPEAT = auto()
    MISSING_CALL = auto()
    NEW_CALL = auto()
    UNKNOWN = auto()

    @classmethod
    def get_random_error(cls) -> ErrorType:
        available_keys = [
            cls.MISSING_PARAMETER,
            cls.MADE_UP_PARAMETER,
            cls.MADE_UP_ASSIGNMENT,
        ]

        return choice(available_keys)



from typing import List, Optional, Tuple
from re import match
from warnings import warn

TOKEN = "var"


def merge_quoted_parameters(list_of_parameters: List[str]) -> List[str]:
    new_parameters = []

    new_parameter = ""
    merge_on = False
    num_mergers = 0

    for parameter in list_of_parameters:
        if parameter.startswith('"') and parameter.count('"') % 2 == 1:
            merge_on = True

        if '="' in parameter and parameter.count('"') % 2 == 1:
            merge_on = True

        if merge_on:
            new_parameter += f'{", " if new_parameter else ""}{parameter}'
            num_mergers += 1

        if new_parameter.count('"') % 2 == 0:
            merge_on = False

        if not merge_on:
            temp = new_parameter or parameter
            new_parameter = ""

            if temp:
                new_parameters.append(temp)

    if num_mergers > 0:
        return merge_quoted_parameters(new_parameters)

    return new_parameters


def parse_parameters(signature: str) -> Tuple[str, List[str]]:
    try:
        match_object = match(
            pattern=r"\s*(?P<action_name>.*)\((?P<parameters>.*)\)\s*",
            string=signature,
        )

        if match_object:
            action_name = match_object.groupdict().get("action_name", "")
            parameters = match_object.groupdict().get("parameters", [])

            parameters = [p.strip() for p in parameters.split(",")]
            parameters = merge_quoted_parameters(parameters)

            return action_name, parameters
        else:
            return "", []

    except Exception as e:
        warn(
            message=f"Could not parse {signature}: {e}", category=SyntaxWarning
        )
        return "", []


def get_token(index: int, token: str = TOKEN) -> str:
    return f"{token}{index}"


def extract_label(label_string: Optional[str]) -> Tuple[str, Optional[str]]:
    label, mapping = "", None

    if label_string is not None:
        if "." in label_string:
            match_object = match(
                pattern=r".*\$(?P<label>[^\.]*)\.(?P<map>.*)\$.*",
                string=label_string,
            )

            if match_object:
                label = match_object.groupdict().get("label", "")
                mapping = match_object.groupdict().get("map", None)

        else:
            match_object = match(
                pattern=r".*\$(?P<map>.*)\$.*",
                string=label_string,
            )

            if match_object:
                label = get_token(index=0)
                mapping = match_object.groupdict().get("map", None)

    # TODO: Generic transformation
    if mapping is not None and "[" in mapping:
        mapping = mapping.split("[")[0]

    return label, mapping



from __future__ import annotations
from pydantic import BaseModel, ConfigDict, Field, model_validator
from typing import Dict, Optional, Any, Union, List
from http import HTTPStatus, HTTPMethod


class OpenAPIInfo(BaseModel):
    title: str
    description: Optional[str] = ""
    version: str


class Parameter(BaseModel):
    name: str
    description: Optional[str] = ""
    required: Optional[bool] = False
    type: Optional[str] = None
    parameter_schema: Optional[Component] = Field(alias="schema", default=None)


class Component(BaseModel):
    title: Optional[str] = ""
    type: Optional[str] = None
    description: Optional[str] = ""
    enum: List[Union[str, Any]] = []
    required: List[str] = []
    properties: Dict[str, Union[ResponseSelection, Component, Any]] = {}
    items: Optional[Component] = None

    def transform_properties_to_parameter(
        self, is_required: Optional[bool] = None
    ) -> List[Parameter]:
        parameters = []

        for key, item in self.properties.items():
            if isinstance(item, Component):
                parameters.append(
                    Parameter(
                        name=key,
                        description=item.description,
                        required=key in self.required or is_required,
                        type=item.type,
                        schema=item,
                    )
                )

            elif isinstance(item, ResponseSelection):
                nested_parameters = item.get_parameters(is_required)
                parameters.append(
                    Parameter(
                        name=key,
                        required=is_required,
                    )
                )

                if nested_parameters:
                    for np in nested_parameters:
                        parameters.append(
                            Parameter(
                                name=f"{key}.{np.name}",
                                description=np.description,
                                required=np.required,
                                type=np.type,
                                schema=np.parameter_schema,
                            )
                        )
                else:
                    parameters.append(Parameter(name=key, required=is_required))
            else:
                pass

        if self.items:
            parameters.extend(
                self.items.transform_properties_to_parameter(is_required)
            )

        return parameters


class ResponseSelection(BaseModel):
    model_config = ConfigDict(extra="ignore")

    oneOf: List[Component] = []
    allOf: List[Component] = []
    anyOf: List[Component] = []

    def get_parameters(
        self, is_required: Optional[bool] = None
    ) -> List[Parameter]:
        for key in self.model_fields.keys():
            list_of_components = getattr(self, key)

            if list_of_components:
                merged_properties: List[Parameter] = []

                for component in list_of_components:
                    merged_properties.extend(
                        component.transform_properties_to_parameter(is_required)
                    )

                return merged_properties

        return []


class SchemaObject(BaseModel):
    object_schema: Union[ResponseSelection, Component] = Field(alias="schema")

    def get_parameters(
        self, is_required: Optional[bool] = None
    ) -> List[Parameter]:
        if isinstance(self.object_schema, Component):
            if self.object_schema.type == "array":
                # TODO: https://github.com/TathagataChakraborti/NESTFUL/issues/4
                items = self.object_schema.items

                if not items:
                    raise ValueError(f"Empty array specification: {self}")
                else:
                    if items.type != "object":
                        raise NotImplementedError

                    return items.transform_properties_to_parameter(is_required)

            elif self.object_schema.type == "object":
                return self.object_schema.transform_properties_to_parameter(
                    is_required
                )

            else:
                # TODO: https://github.com/TathagataChakraborti/NESTFUL/issues/3
                return [Parameter(name="None")]

        elif isinstance(self.object_schema, ResponseSelection):
            return self.object_schema.get_parameters(is_required)

        else:
            raise NotImplementedError


class PayloadObject(BaseModel):
    model_config = ConfigDict(extra="ignore")

    description: Optional[str] = ""
    content: Dict[str, SchemaObject] = {}
    required: Optional[bool] = None

    def get_parameters(
        self, serialization_scheme: Optional[str] = None
    ) -> List[Parameter]:
        output_schema = self.get_output_schema(serialization_scheme)
        return (
            output_schema.get_parameters(is_required=self.required)
            if output_schema
            else []
        )

    def get_output_schema(
        self, serialization_scheme: Optional[str] = None
    ) -> Optional[SchemaObject]:
        if serialization_scheme:
            return self.content.get(serialization_scheme, None)

        else:
            return next(iter(self.content.values()))


class PathSpec(BaseModel):
    model_config = ConfigDict(extra="ignore")

    operationId: str
    description: Optional[str] = ""
    summary: Optional[str] = ""
    parameters: List[Parameter] = []
    requestBody: Optional[PayloadObject] = None
    responses: Dict[HTTPStatus, PayloadObject]

    @model_validator(mode="after")
    def merge_request_body_into_parameters(self) -> PathSpec:
        if self.requestBody:
            request_body_parameters = self.requestBody.get_parameters()
            self.parameters.extend(request_body_parameters)

        return self


class ComponentDict(BaseModel):
    schemas: Dict[str, Component]


class OpenAPI(BaseModel):
    model_config = ConfigDict(extra="ignore")

    openapi: str
    info: OpenAPIInfo
    paths: Dict[str, Dict[str, Union[PathSpec, List[Parameter]]]]
    components: Optional[ComponentDict] = None

    @model_validator(mode="after")
    def merge_global_parameters(self) -> OpenAPI:
        new_paths = {}

        for path, path_object in self.paths.items():
            global_parameters = path_object.get("parameters", [])

            if isinstance(global_parameters, PathSpec):
                pass
            else:
                new_spec: Dict[str, PathSpec] = {}

                for method in HTTPMethod:
                    path_spec = path_object.get(method.lower(), None)

                    if isinstance(path_spec, PathSpec):
                        path_spec.parameters.extend(global_parameters)
                        new_spec[method] = path_spec

                new_paths[path] = new_spec

        self.paths = new_paths  # type: ignore
        return self



from enum import StrEnum, auto
from pathlib import Path
from json import load
from os import listdir
from os.path import isfile, join
from jsonref import replace_refs
from http import HTTPStatus
from typing import Any, Optional, Tuple, List

from nestful import SequencingDataset, SequencingData, Catalog, API
from nestful.schemas.openapi import OpenAPI, Component, PathSpec
from nestful.schemas.api import QueryParameter


class DataType(StrEnum):
    DATA = auto()
    SPEC = auto()


def data_path_constructor(
    data_type: DataType = DataType.DATA,
    executable: bool = True,
    name: Optional[str] = None,
    version: Optional[str] = "v1",
) -> Path:
    assert (
        version == "v1"
    ), "The NESTFUL wrapper only supports version v1 for now."

    path_to_file = Path(__file__).parent.resolve()

    exe_string = f"{'non-' if not executable else ''}executable"
    name = f"-{name}" if name else ""

    relative_path_to_data = (
        f"../data_{version}/{exe_string}/{exe_string}{name}-{data_type}.json"
    )

    return Path.joinpath(path_to_file, relative_path_to_data).resolve()


def read_raw_data(
    data_type: DataType = DataType.DATA,
    executable: bool = True,
    name: Optional[str] = None,
    version: Optional[str] = "v1",
) -> Any:
    path_to_data = data_path_constructor(data_type, executable, name, version)

    with open(path_to_data) as f:
        data = load(f)

    return data


def get_nestful_catalog(
    executable: bool = True,
    name: Optional[str] = None,
    version: Optional[str] = "v1",
) -> Catalog:
    data = read_raw_data(DataType.SPEC, executable, name, version)
    return Catalog(apis=[API.model_validate(item) for item in data])


def get_nestful_data(
    executable: bool = True,
    name: Optional[str] = None,
    version: Optional[str] = "v1",
) -> Tuple[SequencingDataset, Catalog]:
    catalog = get_nestful_catalog(executable, name, version)
    raw_sequence_data = read_raw_data(DataType.DATA, executable, name, version)
    sequence_data = SequencingDataset(data=[])

    for item in raw_sequence_data:
        sequence_instance = SequencingData.model_validate(item)
        sequence_data.data.append(sequence_instance)

    return sequence_data, catalog


def get_nestful_data_instance(
    index: int,
    executable: bool = True,
    name: Optional[str] = None,
    version: Optional[str] = "v1",
) -> Tuple[SequencingData, Catalog]:
    sequence_data, catalog = get_nestful_data(executable, name, version)

    assert index < len(sequence_data.data), (
        f"Requested dataset has {len(sequence_data.data)} samples, asked"
        f" for {index}!"
    )

    return sequence_data.data[index], catalog


def parse_api_from_openapi_spec(openapi_spec: OpenAPI) -> List[API]:
    list_of_apis = []
    path_objects = openapi_spec.paths

    for _, item in path_objects.items():
        for http_method, path_spec in item.items():
            if not isinstance(path_spec, PathSpec):
                continue

            for http_status, response_object in path_spec.responses.items():
                if http_status == HTTPStatus.OK:
                    output_parameters = response_object.get_parameters()

                    new_api = API(
                        name=path_spec.operationId,
                        description=path_spec.description or "",
                        query_parameters={
                            p.name: QueryParameter(required=p.required is True)
                            for p in path_spec.parameters
                        },
                        output_parameters={
                            item.name: Component() for item in output_parameters
                        },
                    )

                    list_of_apis.append(new_api)

    return list_of_apis


def get_catalog_from_openapi_specs(abs_path_to_specs: Path) -> Catalog:
    new_catalog = Catalog()

    if str(abs_path_to_specs).endswith(".json"):
        only_json_files = [str(abs_path_to_specs)]
    else:
        only_json_files = [
            filename
            for filename in listdir(abs_path_to_specs)
            if isfile(join(abs_path_to_specs, filename))
            and filename.endswith(".json")
        ]

    for file in only_json_files:
        try:
            with open(join(abs_path_to_specs, file)) as f:
                openapi_spec = load(f)

            openapi_spec_flat = replace_refs(
                openapi_spec, proxies=False, lazy_load=False
            )

            openapi_object = OpenAPI(**openapi_spec_flat)
            new_catalog.apis.extend(parse_api_from_openapi_spec(openapi_object))

        except Exception as e:
            print(e)

    return new_catalog



from __future__ import annotations
from typing import List, Dict, Optional, Any, Union, Tuple, Set
from pydantic import BaseModel, ConfigDict, model_validator
from nestful.utils import parse_parameters, extract_label
from nestful.schemas.api import Catalog, API, MinifiedAPI
from nestful.schemas.errors import ErrorType
from copy import deepcopy

DUMMY_VALUE = "INIT"


class AtomicCall(BaseModel):
    call: SequenceStep
    memory: Dict[str, Any]
    ground_truth: Optional[AtomicCall] = None


class SequenceStep(BaseModel):
    model_config = ConfigDict(extra="ignore")

    name: Optional[str] = ""
    arguments: Dict[str, Any] = dict()
    label: Optional[str] = None
    errors: List[ErrorTag] = []

    def __str__(self) -> str:
        return str(self.model_dump(exclude={"errors"}))

    def generate_dummy_output(self, catalog: Catalog) -> Dict[str, Any]:
        new_memory: Dict[str, Any] = dict()
        api_spec = next(
            filter(lambda api: api.name == self.name, catalog.apis), None
        )

        if api_spec is None or self.label is None:
            return new_memory

        else:
            for k, item in api_spec.output_parameters.items():
                new_memory[k] = (
                    {key: DUMMY_VALUE for key in item.properties}
                    if item.properties
                    else DUMMY_VALUE
                )

            memory = {self.label: new_memory}
            return memory

    def get_tool_spec(self, catalog: Catalog) -> Optional[API]:
        tool_spec = catalog.get_api(name=self.name or "")

        assert not isinstance(tool_spec, MinifiedAPI)
        return tool_spec

    def get_required_args(self, catalog: Catalog) -> Set[str]:
        api_spec = (
            catalog.get_api(name=self.name, required=True)
            if self.name
            else None
        )

        required_arguments = set()

        if isinstance(api_spec, API):
            for item in self.arguments:
                if item in api_spec.get_arguments(required=True):
                    required_arguments.add(item)

        return required_arguments

    def is_same_as(
        self,
        ground_truth: SequenceStep,
        catalog: Catalog,
        required_schema_only: bool = False,
        check_values: bool = False,
    ) -> bool:
        gt_arguments = (
            ground_truth.get_required_args(catalog)
            if required_schema_only
            else set(ground_truth.arguments.keys())
        )

        self_arguments = (
            self.get_required_args(catalog)
            if required_schema_only
            else set(self.arguments.keys())
        )

        if check_values:
            tmp_1 = {
                k: v
                for k, v in ground_truth.arguments.items()
                if k in gt_arguments
            }
            tmp_2 = {
                k: v for k, v in self.arguments.items() if k in self_arguments
            }

            return self.name == ground_truth.name and tmp_1 == tmp_2

        else:
            return (
                self.name == ground_truth.name
                and gt_arguments == self_arguments
            )

    @model_validator(mode="after")
    def non_string_assignments(self) -> SequenceStep:
        self.arguments = {
            key: str(item) for key, item in self.arguments.items()
        }

        return self

    @staticmethod
    def parse_pretty_print(pretty_print: str) -> SequenceStep:
        split = pretty_print.split(" = ")

        label = split[0] if " = " in pretty_print else ""
        signature = split[0] if len(split) == 1 else split[1]

        action_name, parameters = parse_parameters(signature)

        arguments = {}
        for item in parameters:
            item_split = item.split("=")
            arguments[item_split[0]] = item_split[1].replace('"', "")

        return SequenceStep(name=action_name, arguments=arguments, label=label)

    def pretty_print(
        self,
        mapper_tag: Optional[str] = None,
        collapse_maps: bool = True,
    ) -> str:
        label = f"{self.label} = " if self.label else ""

        required_arguments = list(self.arguments.keys())
        pretty_strings = []

        if collapse_maps:
            required_arguments = [
                f'{item}="{self.arguments.get(item)}"'
                for item in required_arguments
            ]

        else:
            assert (
                mapper_tag
            ), "You must provide a mapper tag if you are not collapsing maps."

            for item in required_arguments:
                value = self.arguments.get(item)

                if item != value:
                    mapping_string = f'{mapper_tag}("{value}", {item})'
                    pretty_strings.append(mapping_string)

        action_string = f"{label}{self.name}({', '.join(required_arguments)})"
        pretty_strings.append(action_string)

        return "\n".join(pretty_strings)

    def remove_reference(self, label: str) -> SequenceStep:
        new_step = deepcopy(self)
        new_step.arguments = dict()

        for arg, value in self.arguments.items():
            l, m = extract_label(str(value))

            if l == label:
                continue

            new_step.arguments[arg] = value

        return new_step


class SequencingData(BaseModel):
    input: str = ""
    output: List[SequenceStep] = []
    var_result: Dict[str, str] = {}
    errors: List[ErrorTag] = []

    @model_validator(mode="after")
    def remove_final_step(self) -> SequencingData:
        if self.output and self.output[-1].name == "var_result":
            self.var_result = self.output[-1].arguments
            self.output = self.output[:-1]

        return self

    def __str__(self) -> str:
        list_of_str = [str(item) for item in self.output]
        string_form = ",\n".join(list_of_str)
        return f"[\n{string_form}\n]"

    def generate_dummy_output(
        self,
        catalog: Catalog,
        index: Optional[int] = None,
    ) -> Dict[str, Any]:
        assert index is None or index < len(self.output)
        index = len(self.output) if index is None else index

        memory: Dict[str, Any] = {}

        for i in range(index):
            step_memory = self.output[i].generate_dummy_output(catalog)
            memory = {**memory, **step_memory}

        return memory

    def get_tool_specs(self, catalog: Catalog) -> List[API]:
        list_of_apis: List[API] = []

        for step in self.output:
            tool_spec = step.get_tool_spec(catalog)

            if isinstance(tool_spec, API) and tool_spec not in list_of_apis:
                list_of_apis.append(tool_spec)

        return list_of_apis

    def contains(
        self,
        step: SequenceStep,
        catalog: Catalog,
        required_schema_only: bool = False,
        check_values: bool = False,
    ) -> bool:
        return any(
            [
                item.is_same_as(
                    step, catalog, required_schema_only, check_values
                )
                for item in self.output
            ]
        )

    def is_same_as(
        self,
        ground_truth: SequencingData,
        catalog: Catalog,
        required_schema_only: bool = False,
        check_values: bool = False,
    ) -> bool:
        return all(
            [
                ground_truth.contains(
                    step, catalog, required_schema_only, check_values
                )
                for step in self.output
            ]
        ) and all(
            [
                self.contains(step, catalog, required_schema_only, check_values)
                for step in ground_truth.output
            ]
        )

    def remove_reference(self, label: str) -> SequencingData:
        new_sequence = deepcopy(self)
        cached_index = 0

        for index, step in enumerate(new_sequence.output):
            new_sequence.output[index] = step.remove_reference(label)

            if step.label == label:
                cached_index = index

        new_sequence.output = (
            new_sequence.output[:cached_index]
            + new_sequence.output[cached_index + 1 :]
        )

        return new_sequence

    def who_used(self, label: str) -> List[int]:
        indices = []

        for index, step in enumerate(self.output):
            for arg, value in step.arguments.items():
                l, m = extract_label(str(value))

                if l == label:
                    indices.append(index)
                    break

        return indices

    def who_produced(self, var: str) -> Tuple[Optional[str], int]:
        index_map: Dict[str, int] = {}

        for step in self.output:
            if step.name is not None:
                current_index = index_map.get(step.name, 0)
                index_map[step.name] = current_index + 1

                if step.label == var:
                    return step.name, index_map[step.name]

        return None, 0

    def get_label(self, name: str, index: int = 1) -> Optional[str]:
        index_map: Dict[str, int] = {}

        for step in self.output:
            if step.name is not None:
                current_index = index_map.get(step.name, 0)
                index_map[step.name] = current_index + 1

                if step.name == name and index_map[step.name] == index:
                    return step.label

        return None

    @staticmethod
    def parse_pretty_print(
        pretty_print: Union[str, List[str]]
    ) -> SequencingData:
        if isinstance(pretty_print, str):
            pretty_print = pretty_print.split("\n")

        return SequencingData(
            input="",
            output=[SequenceStep.parse_pretty_print(p) for p in pretty_print],
        )

    def pretty_print(
        self,
        mapper_tag: Optional[str] = None,
        collapse_maps: bool = True,
    ) -> str:
        tokens = [
            op.pretty_print(mapper_tag, collapse_maps) for op in self.output
        ]

        return "\n".join(tokens)


class SequencingDataset(BaseModel):
    data: List[SequencingData]


class ErrorTag(BaseModel):
    error_type: ErrorType = ErrorType.UNKNOWN
    info: str | Dict[str, Any] | None
